/*
 *	dq12monster.c
 *		Copyright (C)2001-2003 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "filesfc.h"
#include "dq12decode.h"
#include "dq12item.h"
#include "dq12spell.h"
#include "dq12monster.h"

#if defined(MODE_DQ1)
#define FILEPOS_MONNAME_ADDR	0x05b5eb
#define FILEPOS_MONDATA			0x05da0e
#define FILEPOS_MONENC			0x05b52d
#define FILEPOS_MONENC_LV		0x05b591
#define NUM_MON					40
#define NUM_MONENC				19
#define SIZEOF_MONENC			5
#else /* MODE_DQ1 */
#define FILEPOS_MONNAME_ADDR	0x05b63d
#define FILEPOS_MONDATA			0x05dcde
#define FILEPOS_MONENC			0x05bec5
#define FILEPOS_MONENC_LV		0x05b5a5
#define FILEPOS_MONENC_SUB		0x05c0e5
#define NUM_MON					82
#define NUM_MONENC				68
#define NUM_MONENC_SUB			20
#define SIZEOF_MONENC			8
#define SIZEOF_MONENC_SUB		4
#endif /* MODE_DQ1 */

#define FILEPOS_MONNAME_BANK	0x000000
#define FILEPOS_MONMAP			0x05b291
#define SIZEOF_MONDATA			18
#define NUM_MONMAP				4
#define SIZEOF_MONENC_LV		1
#define SIZEOF_OFFSETADDR		2

static void mon_setname(void);

static char tbl_monname[NUM_MON][256];	/* X^[̖Oi[e[u */
static int flag_setmonname = 0;			/* tbl_monnameɃf[^i[Ă邩ǂ */

static const struct
{
	int width;
	int height;
} tbl_sizeof_monmap[NUM_MONMAP] =
{
	{ 7,  6},
	{10,  9},
	{16, 16},
	{16, 16},
};

const char *mon_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MON <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmonname)
		mon_setname();

	return tbl_monname[index];
}

static void mon_setname(void)
{
	int i;
	int tbl_offset[NUM_MON];

	/* X^[̖Of[^ */
	for (i = 0; i < NUM_MON; i++)
		*tbl_monname[i] = '\0';

	/* X^[̖Of[^i[ĂAhX(ւ̃ItZbg)ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_OFFSETADDR];

		file_read_sfc(buf, FILEPOS_MONNAME_ADDR + SIZEOF_OFFSETADDR * i, SIZEOF_OFFSETADDR);
		tbl_offset[i] = (buf[1] << 8) | buf[0];
	}

	/* X^[̖Of[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[256];
		unsigned char *ptr_term = buf;

		file_read_sfc(buf, FILEPOS_MONNAME_BANK + tbl_offset[i], sizeof(buf));

		if ((ptr_term = memchr(buf, 0xff, sizeof(buf))) == NULL)
			break;

		dec_string(tbl_monname[i], buf, sizeof(tbl_monname[i]), ptr_term - buf, 0);
	}

	flag_setmonname = 1;
}

mondata_t *mon_readdata(void)
{
	int i, j;
	mondata_t *buf_mondata = NULL;

	/* X^[̊ef[^i[obt@mۂ */
	if ((buf_mondata = malloc(sizeof(mondata_t) * NUM_MON)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[̊ef[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONDATA];

		file_read_sfc(buf, FILEPOS_MONDATA + SIZEOF_MONDATA * i, SIZEOF_MONDATA);

		/* lnf[^ */
		buf_mondata[i].hit_point = buf[ 3];

#if defined(MODE_DQ2)
		/* ʈ */
		if (i == 0x4f)							/* $0B9C07 */
			buf_mondata[i].hit_point = 320;		/* $0B9C14 */
		else if (i == 0x50)						/* $0B9C0B */
			buf_mondata[i].hit_point = 460;		/* $0B9C19 */
		else if (i == 0x51)						/* $0B9C0F */
			buf_mondata[i].hit_point = 1750;	/* $0B9C1E */
#endif /* MODE_DQ2 */

		buf_mondata[i].magic_power      = buf[ 4];
		buf_mondata[i].experience_point = (buf[17] << 8) | buf[16];
		buf_mondata[i].gold             = ((buf[14] & 0xc0) << 2) | buf[ 5];
		buf_mondata[i].attack_power     = buf[ 1];
		buf_mondata[i].defense_power    = buf[ 2];
		buf_mondata[i].agility          = buf[ 0];

		/* ϐnf[^ */
		buf_mondata[i].regist_damage    = (buf[ 6] & 0xe0) >> 5;
		buf_mondata[i].regist_sleep     = (buf[ 7] & 0xe0) >> 5;
		buf_mondata[i].regist_stopspell = (buf[ 8] & 0xe0) >> 5;
		buf_mondata[i].regist_defeat    = (buf[ 9] & 0xe0) >> 5;
		buf_mondata[i].regist_surround  = (buf[10] & 0xe0) >> 5;
		buf_mondata[i].regist_defense   = (buf[11] & 0xe0) >> 5;

		/* ̑̃f[^ */
		buf_mondata[i].item_no     = buf[15];
		buf_mondata[i].item_get    = (buf[13] & 0x60) >> 5;
		buf_mondata[i].evade_rate  = buf[14] & 0x0f;
		buf_mondata[i].action_type = (buf[12] & 0xe0) >> 5;
		buf_mondata[i].judgment    = (buf[14] & 0x30) >> 4;

		/* sp^[f[^ */
		for (j = 0; j < 8; j++)
			buf_mondata[i].action[j] = buf[j + 6] & 0x1f;
	}

	return buf_mondata;
}

monenc_t *mon_readenc(void)
{
	int i, j;
	monenc_t *buf_monenc = NULL;
#if !defined(MODE_DQ1)
	int tbl_subparty[NUM_MONENC_SUB][4];
#endif /* MODE_DQ1 */

	/* X^[e[uf[^i[obt@mۂ */
	if ((buf_monenc = malloc(sizeof(monenc_t) * NUM_MONENC)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

#if defined(MODE_DQ1)
	/* X^[e[uf[^ǂݍ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		/* oX^[ */
		{
			unsigned char buf[SIZEOF_MONENC];

			file_read_sfc(buf, FILEPOS_MONENC + SIZEOF_MONENC * i, SIZEOF_MONENC);

			for (j = 0; j < 5; j++)
				buf_monenc[i].monster_no[j] = (buf[j] != 0xff) ? buf[j] - 1 : -1;
		}

		/* X^[p[eBx */
		{
			unsigned char byte;

			file_read_sfc(&byte, FILEPOS_MONENC_LV + SIZEOF_MONENC_LV * i, SIZEOF_MONENC_LV);

			buf_monenc[i].level = byte;
		}
	}
#else /* MODE_DQ1 */
	/* X^[Tup[eBf[^ǂݍ */
	for (i = 0; i < NUM_MONENC_SUB; i++)
	{
		unsigned char buf[SIZEOF_MONENC_SUB];

		file_read_sfc(buf, FILEPOS_MONENC_SUB + SIZEOF_MONENC_SUB * i, SIZEOF_MONENC_SUB);

		for (j = 0; j < 4; j++)
			tbl_subparty[i][j] = (buf[j] != 0xff) ? buf[j] - 1 : -1;
	}

	/* X^[e[uf[^ǂݍ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		/* oX^[ETup[eB */
		{
			unsigned char buf[SIZEOF_MONENC];

			file_read_sfc(buf, FILEPOS_MONENC + SIZEOF_MONENC * i, SIZEOF_MONENC);

			for (j = 0; j < 6; j++)
				buf_monenc[i].monster_no[j] = (buf[j] != 0xff) ? buf[j] - 1 : -1;

			if (buf[6] == 0xff)
			{
				for (j = 0; j < 4; j++)
					buf_monenc[i].subparty[j] = -1;
			}
			else
			{
				if (buf[6] >= NUM_MONENC_SUB)
					break;

				for (j = 0; j < 4; j++)
					buf_monenc[i].subparty[j] = tbl_subparty[buf[6]][j];
			}
		}

		/* X^[p[eBx */
		{
			unsigned char byte;

			file_read_sfc(&byte, FILEPOS_MONENC_LV + SIZEOF_MONENC_LV * i, SIZEOF_MONENC_LV);

			buf_monenc[i].level = byte;
		}
	}
#endif /* MODE_DQ1 */

	return buf_monenc;
}

monmap_t *mon_readmap(void)
{
	int i, x, y;
	unsigned char *buf = NULL;
	unsigned char *ptr_buf = NULL;
	monmap_t *buf_monmap = NULL;
	const size_t bufsize = 1024;

	/* X^[}bvf[^i[obt@mۂ */
	if ((buf_monmap = malloc(sizeof(monmap_t) * NUM_MONMAP)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (y = 0; y < 16; y++)
		{
			for (x = 0; x < 16; x++)
				buf_monmap[i].area[x][y] = 0;
		}
	}

	/* X^[}bvf[^ǂݍ */
	if ((buf = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	memset(buf, 0x00, bufsize);
	file_read_sfc(buf, FILEPOS_MONMAP, bufsize);

	ptr_buf = buf;
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (y = 0; y < tbl_sizeof_monmap[i].height; y++)
		{
			for (x = 0; x < tbl_sizeof_monmap[i].width; x++)
				buf_monmap[i].area[x][y] = *ptr_buf++;
		}
	}

	free(buf);
	buf = NULL;

	return buf_monmap;
}

int mon_writeparam(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̐lnf[^o͂ */
	fprintf(fp, "No.      O             "
				"    HP     MP     Ex      G "
				"U  f\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%02X): %-14s :  %5d, %5d, %5ld, %5d, %5d, %5d, %5d\n", i, i,
				mon_getname(i),
				ptr_mondata[i].hit_point,
				ptr_mondata[i].magic_power,
				ptr_mondata[i].experience_point,
				ptr_mondata[i].gold,
				ptr_mondata[i].attack_power,
				ptr_mondata[i].defense_power,
				ptr_mondata[i].agility);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeregist(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̑ϐnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̑ϐnf[^o͂ */
	fprintf(fp, "No.      O             "
				"U Ö  V f 猸\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%02X): %-14s :  %3d, %3d, %3d, %3d, %3d, %3d\n", i, i,
				mon_getname(i),
				ptr_mondata[i].regist_damage,
				ptr_mondata[i].regist_sleep,
				ptr_mondata[i].regist_stopspell,
				ptr_mondata[i].regist_defeat,
				ptr_mondata[i].regist_surround,
				ptr_mondata[i].regist_defense);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeitem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̕󔠂̓f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̕󔠂̓f[^o͂ */
	fprintf(fp, "No.      O             "
				"x   \n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%02X): %-14s :  %3d,  %-14s\n", i, i,
				mon_getname(i),
				ptr_mondata[i].item_get,
				item_getname(ptr_mondata[i].item_no));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeact(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̍sp^[f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̍sp^[f[^o͂ */
	for (i = 0; i < NUM_MON; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%02X): %s]\n", i, i, mon_getname(i));

		fprintf(fp, "No. s\n");
		for (j = 0; j < 8; j++)
		{
			const char *ptr_spellname = spell_getactname(ptr_mondata[i].action[j]);
			const char *ptr_spelldesc = spell_getactdesc(ptr_mondata[i].action[j]);

			fprintf(fp, "%02d:", j + 1);
			if (strlen(ptr_spellname) > 0)
				fprintf(fp, " %s", ptr_spellname);
			if (strlen(ptr_spelldesc) > 0)
				fprintf(fp, " (%s)", ptr_spelldesc);
			fputc('\n', fp);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemisc(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̂̑̃f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̂̑̃f[^o͂ */
	fprintf(fp, "No.      O             "
				" f\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%02X): %-14s :  %3d, %3d\n", i, i,
				mon_getname(i),
				ptr_mondata[i].evade_rate,
				ptr_mondata[i].judgment);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeenc(const char *ptr_filename, const monenc_t *ptr_monenc)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "--------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monenc == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[e[uf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

#if defined(MODE_DQ1)
	/* X^[e[uo͂ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[%03d(%02X)]\n", i, i);
		fprintf(fp, "x: %3d\n", ptr_monenc[i].level);
		fprintf(fp, "No. X^[\n");
		for (j = 0; j < 5; j++)
		{
			fprintf(fp, "%02d: %-14s\n", j,
					(ptr_monenc[i].monster_no[j] >= 0) ? mon_getname(ptr_monenc[i].monster_no[j]) : str_nothing);
		}
	}
#else /* MODE_DQ1 */
	/* X^[e[uo͂ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[%03d(%02X)]\n", i, i);
		fprintf(fp, "x: %3d\n", ptr_monenc[i].level);
		fprintf(fp, "No. X^[\n");
		for (j = 0; j < 6; j++)
		{
			fprintf(fp, "%02d: %-14s\n", j,
					(ptr_monenc[i].monster_no[j] >= 0) ? mon_getname(ptr_monenc[i].monster_no[j]) : str_nothing);
		}
		fprintf(fp, "Tup[eB: ");
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");

			fprintf(fp, "%-14s", (ptr_monenc[i].subparty[j] >= 0) ? mon_getname(ptr_monenc[i].subparty[j]) : str_nothing);
		}
		fputc('\n', fp);
	}
#endif /* MODE_DQ1 */

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemap(const char *ptr_filename, const monmap_t *ptr_monmap)
{
	int i, x, y;
	FILE *fp = NULL;
	const char *str_name[/* NUM_MONMAP */] =
	{
		"_LA(II)",
		"tB[h(I)",
		"tB[h(II)",
		"C(II)",
	};

	assert(sizeof(str_name) / sizeof(str_name[0]) == NUM_MONMAP);

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monmap == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[}bvf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[}bvo͂ */
	for (i = 0; i < NUM_MONMAP; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%1d: %s]\n", i, str_name[i]);
		for (y = 0; y < tbl_sizeof_monmap[i].height; y++)
		{
			for (x = 0; x < tbl_sizeof_monmap[i].width; x++)
			{
				if (x != 0)
					fprintf(fp, " ");

				if (ptr_monmap[i].area[x][y] != 0xff)
					fprintf(fp, "%03d", ptr_monmap[i].area[x][y]);
				else
					fprintf(fp, "---");
			}
			fputc('\n', fp);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
