/*
 *	dq3analyzer.c
 *		Copyright (C)2001-2003 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "errmes.h"
#include "utility.h"
#include "filesfc.h"
#include "dq3monster.h"
#include "dq3character.h"
#include "dq3item.h"
#include "dq3spell.h"

#define FILENAME_MON_PARAM		"dq3monparam.txt"
#define FILENAME_MON_REGIST		"dq3monregist.txt"
#define FILENAME_MON_FLAG		"dq3monflag.txt"
#define FILENAME_MON_ITEM		"dq3monitem.txt"
#define FILENAME_MON_ACT		"dq3monact.txt"
#define FILENAME_MON_MISC		"dq3monmisc.txt"
#define FILENAME_MON_ENC		"dq3monenc.txt"
#define FILENAME_MON_MAP		"dq3monmap.txt"
#define FILENAME_CHR_LVUP		"dq3chrlvup.txt"
#define FILENAME_ITEM_PARAM		"dq3itemparam.txt"
#define FILENAME_SPELL_PARAM	"dq3spellparam.txt"

#define STR_TOO_FEW_ARGUMENTS		"Too few arguments"
#define STR_INVALID_OPTION			"Invalid option"
#define STR_INVALID_ROMIMAGE_FILE	"Invalid ROM-Image file"
#define STR_CANNOT_CREATE_FILE		"Cannot create file"

#define STR_TITLE	"DRAGONQUEST3         "
			/* 0x44, 0x52, 0x41, 0x47, 0x4f, 0x4e, 0x51, 0x55, 0x45, 0x53, 0x54, 0x33, 0x20, 0x20, 0x20, 0x20, */
			/* 0x20, 0x20, 0x20, 0x20, 0x20 */

/* 쌠\L */
static const char *str_copyright = "dq3analyzer - Copyright (C)2001, 2002 Shingo Endo <s-endo@yk.rim.or.jp>";

int main(int argc, char *argv[])
{
	int i;
	int ret = EXIT_SUCCESS;
	int flag_invalid = 0;
	int flag_usage   = 0;
	char str_srcfilename[1024] = "";
	char str_dstpath[1024]     = "";
	char str_dstfilename[1024] = "";
	mondata_t   *buf_mondata   = NULL;
	monenc_t    *buf_monenc    = NULL;
	monmap_t    *buf_monmap    = NULL;
	chrlvup_t   *buf_chrlvup   = NULL;
	itemdata_t  *buf_itemdata  = NULL;
	spelldata_t *buf_spelldata = NULL;

	errno = 0;
	err_setprogname(argv[0]);

	/* R}hC͂ */
	for (i = 1; i < argc; i++)
	{
		if (argv[i] == NULL)
			continue;

		if (*argv[i] == '-')
		{
			const char *ptr = argv[i];

			while (*++ptr != '\0')
			{
				switch (tolower(*ptr))
				{
				case 'h':	/* gp@̕\ */
					flag_usage = 1;
					break;
				default:
					flag_invalid = 1;
					break;
				}
			}
		}
		else
		{
			if (*str_srcfilename == '\0')
				transfilename(str_srcfilename, argv[i], sizeof(str_srcfilename));
			else if (*str_dstpath == '\0')
				transpath(str_dstpath, argv[i], sizeof(str_dstpath));
		}
	}

	/* gp@\ */
	if (flag_usage)
	{
		puts(str_copyright);
		printf("%s: [option] src-filename [dst-pathname]\n", err_getprogname());
		printf("\t-h\tdisplay this help and exit\n");
		printf("\n");

		exit(EXIT_SUCCESS);
	}

	/* R}hC`FbN */
	if (flag_invalid)
		err_printf(STR_INVALID_OPTION);
	if (*str_srcfilename == '\0')
		err_printf(STR_TOO_FEW_ARGUMENTS);

	/* t@CI[v */
	if (!file_open_sfc(str_srcfilename, STR_TITLE))
		err_printf(STR_INVALID_ROMIMAGE_FILE " - %s", str_srcfilename);

	/* ef[^́E͂ */
	buf_mondata   = mon_readdata();
	buf_monenc    = mon_readenc();
	buf_monmap    = mon_readmap();
	buf_chrlvup   = chr_readlvup();
	buf_itemdata  = item_readdata();
	buf_spelldata = spell_readdata();

	/* ef[^o͂ */
	if (buf_mondata   == NULL ||
		buf_monenc    == NULL ||
		buf_monmap    == NULL ||
		buf_chrlvup   == NULL ||
		buf_itemdata  == NULL ||
		buf_spelldata == NULL)
	{
		wrn_printf(STR_INVALID_ROMIMAGE_FILE " - %s*", str_srcfilename);
		ret = EXIT_FAILURE;
	}
	else
	{
		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_PARAM);
		if (!mon_writeparam(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_REGIST);
		if (!mon_writeregist(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_ITEM);
		if (!mon_writeitem(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_ACT);
		if (!mon_writeact(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_MISC);
		if (!mon_writemisc(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_ENC);
		if (!mon_writeenc(str_dstfilename, buf_monenc))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_MAP);
		if (!mon_writemap(str_dstfilename, buf_monmap))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}
#if 0
		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_CHR_LVUP);
		if (!chr_writelvup(str_dstfilename, buf_chrlvup))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}
#endif
		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_PARAM);
		if (!item_writeparam(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_SPELL_PARAM);
		if (!spell_writeparam(str_dstfilename, buf_spelldata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}
	}

	/* t@CN[Y */
	file_close_sfc();

	/* obt@ */
	free(buf_mondata);
	free(buf_monenc);
	free(buf_monmap);
	free(buf_chrlvup);
	free(buf_itemdata);
	free(buf_spelldata);
	buf_mondata   = NULL;
	buf_monenc    = NULL;
	buf_monmap    = NULL;
	buf_chrlvup   = NULL;
	buf_itemdata  = NULL;
	buf_spelldata = NULL;

	return ret;
}
