/*
 *	dq3monster.c
 *		Copyright (C)2001-2003 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "filesfc.h"
#include "dq3define.h"
#include "dq3item.h"
#include "dq3spell.h"
#include "dq3monster.h"

#define FILEPOS_MONDATA		0x020000
#define FILEPOS_MONENC		0x08add1
#define FILEPOS_MONMAP		0x08b7fd
#define NUM_MON				160
#define NUM_MONENC			105
#define NUM_MONMAP			6
#define SIZEOF_MONDATA		37
#define SIZEOF_MONENC		20

static void mon_setnameindex(void);

static int tbl_monnameindex[NUM_MON];	/* X^[\̃CfbNXi[e[u */
static int flag_monnameindex = 0;		/* tbl_monnameindexɃf[^i[Ă邩ǂ */

const char *mon_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MON <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_monnameindex)
		mon_setnameindex();

	return str_getdef(tbl_monnameindex[index]);
}

static void mon_setnameindex(void)
{
	int i;

	/* X^[̃CfbNXǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONDATA];

		file_read_sfc(buf, FILEPOS_MONDATA + SIZEOF_MONDATA * i, SIZEOF_MONDATA);
		tbl_monnameindex[i] = (buf[ 1] << 8) | buf[ 0];
	}

	flag_monnameindex = 1;
}

mondata_t *mon_readdata(void)
{
	int i;
	mondata_t *buf_mondata = NULL;

	/* X^[̊ef[^i[obt@mۂ */
	if ((buf_mondata = malloc(sizeof(mondata_t) * NUM_MON)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[̊ef[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONDATA];

		file_read_sfc(buf, FILEPOS_MONDATA + SIZEOF_MONDATA * i, SIZEOF_MONDATA);

		/* lnf[^ */
		buf_mondata[i].name_index       = (buf[ 1] << 8) | buf[ 0];
		buf_mondata[i].level            = buf[ 2] & 0x3f;
		buf_mondata[i].hit_point        = (buf[35] << 8) | buf[34];
		buf_mondata[i].magic_power      = buf[11];
		buf_mondata[i].experience_point = (buf[ 4] << 8) | buf[ 3];
		buf_mondata[i].gold             = ((buf[ 6] & 0x03) << 8) | buf[ 5];
		buf_mondata[i].attack_power     = ((buf[ 7] & 0x0f) << 6) | ((buf[ 6] & 0xfc) >> 2);
		buf_mondata[i].defense_power    = ((buf[ 8] & 0x3f) << 4) | ((buf[ 7] & 0xf0) >> 4);
		buf_mondata[i].agility          = buf[10];

		/* ϐnf[^ */
		buf_mondata[i].regist_fire      = (buf[30] & 0x30) >> 4;
		buf_mondata[i].regist_ice       = (buf[30] & 0xc0) >> 6;
		buf_mondata[i].regist_infernos  = buf[31] & 0x03;
		buf_mondata[i].regist_lightning = (buf[31] & 0x0c) >> 2;
		buf_mondata[i].regist_defeat    = (buf[31] & 0x30) >> 4;
		buf_mondata[i].regist_sacrifice = (buf[31] & 0xc0) >> 6;
		buf_mondata[i].regist_sleep     = buf[32] & 0x03;
		buf_mondata[i].regist_stopspell = (buf[32] & 0x0c) >> 2;
		buf_mondata[i].regist_defense   = (buf[32] & 0x30) >> 4;
		buf_mondata[i].regist_surround  = (buf[32] & 0xc0) >> 6;
		buf_mondata[i].regist_robmagic  = buf[33] & 0x03;
		buf_mondata[i].regist_chaos     = (buf[33] & 0x0c) >> 2;
		buf_mondata[i].regist_slow      = (buf[33] & 0x30) >> 4;
		buf_mondata[i].regist_expell    = (buf[33] & 0xc0) >> 6;

		/* tOnf[^ */
		buf_mondata[i].flag_type_dragon = ((buf[ 2] & 0x40) != 0x00);
		buf_mondata[i].flag_type_zombie = ((buf[ 2] & 0x80) != 0x00);
		buf_mondata[i].flag_intensive   = ((buf[30] & 0x04) != 0x00);

		/* ̑̃f[^ */
		buf_mondata[i].item_no           = buf[12];
		buf_mondata[i].item_get          = (buf[29] & 0x38) >> 3;
		buf_mondata[i].evade_rate        = buf[29] & 0x07;
		buf_mondata[i].auto_healing      = buf[30] & 0x03;
		buf_mondata[i].repeatedly_action = (buf[29] & 0xc0) >> 6;
		buf_mondata[i].action_type       = (buf[28] & 0x30) >> 4;
		buf_mondata[i].judgment          = (buf[28] & 0xc0) >> 6;

		/* sp^[f[^ */
		buf_mondata[i].action[0] = ((buf[15] & 0x01) << 8) | ((buf[14] & 0xff) >> 0);
		buf_mondata[i].action[1] = ((buf[16] & 0x03) << 7) | ((buf[15] & 0xfe) >> 1);
		buf_mondata[i].action[2] = ((buf[17] & 0x07) << 6) | ((buf[16] & 0xfc) >> 2);
		buf_mondata[i].action[3] = ((buf[18] & 0x0f) << 5) | ((buf[17] & 0xf8) >> 3);
		buf_mondata[i].action[4] = ((buf[19] & 0x1f) << 4) | ((buf[18] & 0xf0) >> 4);
		buf_mondata[i].action[5] = ((buf[20] & 0x3f) << 3) | ((buf[19] & 0xe0) >> 5);
		buf_mondata[i].action[6] = ((buf[21] & 0x7f) << 2) | ((buf[20] & 0xc0) >> 6);
		buf_mondata[i].action[7] = ((buf[22] & 0xff) << 1) | ((buf[21] & 0x80) >> 7);
	}

	return buf_mondata;
}

monenc_t *mon_readenc(void)
{
	int i, j;
	monenc_t *buf_monenc = NULL;

	/* X^[e[uf[^i[obt@mۂ */
	if ((buf_monenc = malloc(sizeof(monenc_t) * NUM_MONENC)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[e[uf[^ǂݍ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		unsigned char buf[SIZEOF_MONENC];

		file_read_sfc(buf, FILEPOS_MONENC + SIZEOF_MONENC * i, SIZEOF_MONENC);

		buf_monenc[i].level = buf[18];

		for (j = 0; j < 12; j++)
			buf_monenc[i].mon[j].monster_no = buf[j + 4];
	}

	return buf_monenc;
}

monmap_t *mon_readmap(void)
{
	int i, x, y;
	unsigned char *buf = NULL;
	unsigned char *ptr_buf = NULL;
	monmap_t *buf_monmap = NULL;
	const size_t bufsize = 1536;

	/* X^[}bvf[^i[obt@mۂ */
	if ((buf_monmap = malloc(sizeof(monmap_t) * NUM_MONMAP)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[}bvf[^ǂݍ */
	if ((buf = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	memset(buf, 0x00, bufsize);
	file_read_sfc(buf, FILEPOS_MONMAP, bufsize);

	ptr_buf = buf;
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (y = 0; y < 16; y++)
		{
			for (x = 0; x < 16; x++)
				buf_monmap[i].area[x][y] = *ptr_buf++;
		}
	}

	free(buf);
	buf = NULL;

	return buf_monmap;
}

int mon_writeparam(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̐lnf[^o͂ */
	fprintf(fp, "No.      O                "
				" Lv     HP     MP     Ex      G "
				"U  f\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %3d, %5d, %5d, %5ld, %5d, %5d, %5d, %5d\n", i, i,
				mon_getname(i),
				ptr_mondata[i].level,
				ptr_mondata[i].hit_point,
				ptr_mondata[i].magic_power,
				ptr_mondata[i].experience_point,
				ptr_mondata[i].gold,
				ptr_mondata[i].attack_power,
				ptr_mondata[i].defense_power,
				ptr_mondata[i].agility);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeregist(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̑ϐnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̑ϐnf[^o͂ */
	fprintf(fp, "No.      O               "
				"n C ^ d V  Ö  猸 f MPz   ގU\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d\n", i, i,
				mon_getname(i),
				ptr_mondata[i].regist_fire,
				ptr_mondata[i].regist_ice,
				ptr_mondata[i].regist_infernos,
				ptr_mondata[i].regist_lightning,
				ptr_mondata[i].regist_defeat,
				ptr_mondata[i].regist_sacrifice,
				ptr_mondata[i].regist_sleep,
				ptr_mondata[i].regist_stopspell,
				ptr_mondata[i].regist_defense,
				ptr_mondata[i].regist_surround,
				ptr_mondata[i].regist_robmagic,
				ptr_mondata[i].regist_chaos,
				ptr_mondata[i].regist_slow,
				ptr_mondata[i].regist_expell);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeitem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;
	const char *str_nothing = "------------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̕󔠂̓f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̕󔠂̓f[^o͂ */
	fprintf(fp, "No.      O               "
				"x   \n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %3d,  %-18s\n", i, i,
				mon_getname(i),
				ptr_mondata[i].item_get,
				(ptr_mondata[i].item_no != 0x00) ? item_getname(ptr_mondata[i].item_no) : str_nothing);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeact(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_repeat[] =
	{
		"1",
		"1`2",
		"1`3",
		"2",
	};

	assert(sizeof(str_repeat) / sizeof(str_repeat[0]) == 4);

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̍sp^[f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̍sp^[f[^o͂ */
	for (i = 0; i < NUM_MON; i++)
	{
		if (ptr_mondata[i].repeatedly_action >= 4)
			break;

		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%02X): %s]\n", i, i, mon_getname(i));
		fprintf(fp, "s:   %s\n", str_repeat[ptr_mondata[i].repeatedly_action]);

		fprintf(fp, "No. s\n");
		for (j = 0; j < 8; j++)
		{
			const char *ptr_spellname = spell_getname(ptr_mondata[i].action[j]);
			const char *ptr_spelldesc = spell_getdesc(ptr_mondata[i].action[j]);

			fprintf(fp, "%02d:", j + 1);
			if (strlen(ptr_spellname) > 0)
				fprintf(fp, " %s", ptr_spellname);
			if (strlen(ptr_spelldesc) > 0)
				fprintf(fp, " (%s)", ptr_spelldesc);
			fputc('\n', fp);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemisc(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̂̑̃f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̂̑̃f[^o͂ */
	fprintf(fp, "No.      O               "
				"   s W   f\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %2s   %2s   %2s   %3d, %3d, %3d\n", i, i,
				mon_getname(i),
				str_getflag(ptr_mondata[i].flag_type_dragon),
				str_getflag(ptr_mondata[i].flag_type_zombie),
				str_getflag(ptr_mondata[i].flag_intensive),
				ptr_mondata[i].evade_rate,
				ptr_mondata[i].auto_healing,
				ptr_mondata[i].judgment);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeenc(const char *ptr_filename, const monenc_t *ptr_monenc)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monenc == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[e[uf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[e[uo͂ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[%03d(%02X)]\n", i, i);
		fprintf(fp, "x: %3d\n", ptr_monenc[i].level);
		fprintf(fp, "No. X^[\n");
		for (j = 0; j < 12; j++)
		{
			fprintf(fp, "%02d: %-16s\n", j,
					(ptr_monenc[i].mon[j].monster_no != 0) ? mon_getname(ptr_monenc[i].mon[j].monster_no) : str_nothing);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemap(const char *ptr_filename, const monmap_t *ptr_monmap)
{
	int i, x, y;
	FILE *fp = NULL;
	const char *str_name[/* NUM_MONMAP */] =
	{
		"tB[h",
		"AtKh",
		"_~[1",
		"_~[2",
		"tB[h(C)",
		"AtKh(C)",
	};

	assert(sizeof(str_name) / sizeof(str_name[0]) == NUM_MONMAP);

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monmap == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[}bvf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[}bvo͂ */
	for (i = 0; i < NUM_MONMAP; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%1d: %s]\n", i, str_name[i]);
		for (y = 0; y < 16; y++)
		{
			for (x = 0; x < 16; x++)
			{
				if (x != 0)
					fprintf(fp, " ");

				if (ptr_monmap[i].area[x][y] != 0x00)
					fprintf(fp, "%03d", ptr_monmap[i].area[x][y]);
				else
					fprintf(fp, "---");
			}
			fputc('\n', fp);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
